# -*- coding: iso8859-1 -*-
'''
by: Fabius Astelix @2009-02-05
version: @2009-05-05
credits: credits also go to all the wonderful an talented people of Panda3D community expecially Dave (drwr), yijn_jo, pro_rsoft, treeform, ThomasEgi, and countless other fellas being of constant inspiration and help to the whole community.
license: Free for everybody
summary: This is s little p3d framework with some helpful midgets to make life easier to who want to develop P3D projects using the awesome wxWindows library for python.

DOGMA:
1) To use a common P3d script you must make it Pod-compliant; see "How To Make A P3D Script Pod-Compliant" in the README file
2) it is provided a default .prc file ./config/p3dwxpod.prc with the default p3d presets and if there is need of custom presets, they may be defined in a ./config/<appname>.prc file.
---------
\fixme pulire i prc caricati nei pea
\warning @1.5.4 there are issues with makeTextureBuffer so the behaviour of the pod after loaded peas that use texturebuffer may be sloppy
'''
__all__ = ['P3dWxApp', 'P3dWxWindow']
import os, sys, time, wx
#for fix a nasty bug which make loose keyboard control on p3d surface
if wx.Platform == "__WXMSW__": from ctypes import *
#adds shared modules and scripts to the pythonpath
sys.path.append(
  os.path.join(os.getcwd(), 'shared', 'scripts')
)
#
import wxutils as wxut
import p3dutils as p3dut
#to use for conditional statements to execute p3d scripts within the pod, e.g.: if not '_P3DPOD_' in globals(): run()
_P3DPOD_=True
_P3DPOD_DEBUG=True
#===================================================================
# CLASSES
#===================================================================
#
class P3dWxWindow(wx.Frame):
  '''used by the pod as a surface to show the panda3d window'''
  frameWidth      = 450
  frameHeight     = 320
  appinfo=[]
  #------------------------------------------------------------
  #
  def __init__(self, *args, **kw):
    if not kw.get('title'):
        kw['title'] = 'Generic P3d-Wx Shell'
    if not kw.get('size'):
        kw['size'] = wx.Size(self.frameWidth, self.frameHeight)
    wx.Frame.__init__(self, None, -1, *args, **kw)
    self.wp3dsurface=None
    self.menuBar = wx.MenuBar()
    self.SetMenuBar(self.menuBar)
    self.CreateStatusBar(3)
    self.GetStatusBar().SetStatusWidths([-1,50,90])
  #------------------------------------------------------------
  #
  def onGuiInit(self):
    '''override to define a different p3d surface than the whole wx frame window
    '''
    self.wp3dsurface=self
  #------------------------------------------------------------
  #
  def getP3DSurface(self): return self.wp3dsurface.GetHandle()
  def getP3DSurfaceSize(self): return self.wp3dsurface.GetClientSizeTuple()
#===================================================================
#
#\todo rendere autonomi i peas (v.toycar) e quando sara' usare il secondo import di pandac
from pandac.PandaModules import WindowProperties, loadPrcFile, ConfigVariableSearchPath, ConfigVariableString, loadPrcFileData, AntialiasAttrib
#this will load the default prc configuration for the pod.
loadPrcFile("config/p3dwxpod.prc")
#
import direct.directbase.DirectStart
class P3dWxApp(wx.App):
  '''the main pod app'''
  WIDTH=640
  HEIGHT=480
  APPINFO={
    'appname'     : 'Generic Wx pod for Panda3D',
    'appversion'  : "wip.2009-05-05",
    'appsummary'  : "A little p3d framework with some helpful midgets.",
    'copyright'   : "(c) 2009 Fabius Astelix",
    'license'     : "Free for everybody",
    'website'     : ("http://www.panda3d.org/phpbb2/viewtopic.php?t=5359", "Project homepage"),
    'devs'        : ["Fabius Astelix",],
  }
  ID_FILE_OPENSCPT=810
  ID_CAMERA_FPS=901
  ID_CAMERA_AA=902
  ID_CAMERA_WF=903
  ID_CAMERA_TX=904
  #------------------------------------------------------------
  #
  def OnInit(self):
    #prepare and start the p3d-wx hybrid-engine mainloop
    self.wxevt_loop = wx.EventLoop()
    self.wxevt_old_loop = wx.EventLoop.GetActive()
    wx.EventLoop.SetActive(self.wxevt_loop)
    base.taskMgr.add(self._mainLoop, "MainLoopTask")
    #instantiate the wx UI object
    aname=os.path.basename(os.path.splitext(sys.argv[0])[0])
    if os.path.isfile("config/%s.prc"%aname):
      loadPrcFile("config/%s.prc"%aname)
    self.win = self.init_start(
      title=self.APPINFO['appname'] or aname,
      size=wx.Size(self.WIDTH, self.HEIGHT),
    )
    #initialize the UI
    self.SetTopWindow(self.win)
    self.win.onGuiInit()
    self._createMenuBar()
    self.win.SetStatusText("Welcome to %s!"%aname)
    # app icon
    if wx.Platform == "__WXMSW__": appicon="p3d_icon.ico"
    else: appicon="p3d_icon.png"
    _icon = wx.EmptyIcon()
    _icon.LoadFile("shared/data/pictures/icons/"+appicon, wx.BITMAP_TYPE_ANY)
    self.win.SetIcon(_icon)
    #show the wx window
    self.win.Show(True)
    # is essential to let make up wx window before P3D stuff
    self._mainLoop()
    #bind wx events
    self.win.Bind(wx.EVT_SIZE, self.onSize)
    self.win.Bind(wx.EVT_CLOSE, self.onClose)
    self.vetoActivate=False
    self.win.Bind(wx.EVT_ACTIVATE, self.onActivate)
    ###self.win.Bind(wx.EVT_ENTER_WINDOW, self.onEnter)
    ###self.win.Bind(wx.EVT_LEAVE_WINDOW, lambda e: self._EVTEST(e,">>>EVT_LEAVE_WINDOW"))
    #finalize wx startup and start p3d stuff
    self.init_end()
    # time to start p3d window
    #\warning the p3d surface MUST be visible NOW in the wx ui
    while self.wxevt_loop.Pending(): self.wxevt_loop.Dispatch()
    self.ProcessIdle()
    self._startP3D()
    #
    return True
  #------------------------------------------------------------
  #
  def init_start(self, title, size):
    # override this to provide a custom made window
    return P3dWxWindow(
      title=self.APPINFO['appname'] or title,
      size=wx.Size(self.WIDTH, self.HEIGHT),
    )
  #------------------------------------------------------------
  #
  def init_end(self):
    '''override this to perform a finalization init action (e.g. close a starting splash screen)
    '''
    pass
  #------------------------------------------------------------
  #
  def _startP3D(self):
    assert self._open_p3dw() == True
    base.accept("f12", self.onScreenshot)
    #user app overrided method will be called here
    self.onStartP3D()
    #
    # Building up the pod resource manager
    sg=sys.modules['__main__'].__builtins__.globals()
    self.pea_agent=[
      None, p3dut.PwxResmanager(sg, self._open_p3dw, debug=_P3DPOD_DEBUG)
    ]
    self.resman=self.pea_agent[1]
    #\warning from now on every new stuff created will be possibly destroyed
    # Try to autoload a script from command line
    self.autoLoadScript()
  #------------------------------------------------------------
  #
  def _mainLoop(self, task = None):
    '''P3D main loop - here will be dispatched also wx events'''
    if self.wxevt_loop.Pending(): self.wxevt_loop.Dispatch()
    #\note this is a must-have to have onSize() working
    self.ProcessIdle()
    #\note looks like the kb lost issue depends on the arrow-up and arrow-left keys so this is not the right cure
    ###if wx.Platform == "__WXMSW__": self._kbsync()
    if task != None: return task.cont
  #------------------------------------------------------------
  #
  def _kbsync(self):
    '''just to fix a Windows issue that make loose the keyboard focus on the p3d surface
    '''
    if wx.Platform == "__WXMSW__":
      windll.user32.SetFocus(
        windll.user32.GetWindow(self.win.getP3DSurface(),5)
      )
  #------------------------------------------------------------
  #
  def _open_p3dw(self,x=None):
    '''P3D surface initializations'''
    p3dsurface=self.win.getP3DSurface()
    if not p3dsurface:
      print "p3dsurface open failed!"
      return False
    else:
      wp=WindowProperties().getDefault()
      #if there is already an open window it will retrieve the old values
      if base.win:
        wp0=base.win.getProperties()
        if wp0.getOpen():
          wp=WindowProperties()
          wp.addProperties(wp0)
      wp.setUndecorated(True)
      wp.setOpen(True)
      wp.setParentWindow(p3dsurface)
      wp.setOrigin(0,0)
      wp.setForeground(True)
      wp.setSize(*self.win.getP3DSurfaceSize())
      print ">>>opening p3dsurface"
      return base.openDefaultWindow(props=wp)
  #------------------------------------------------------------
  #
  def onSize(self, event=None):
    '''Important method to resize P3d Surface accordingly to his wx window container and to gain keyboard focus
    '''
    if base.win:
      #empty all the wx event queue first
      while self.wxevt_loop.Pending(): self.wxevt_loop.Dispatch()
      wp0=base.win.getProperties()
      if not wp0.getOpen():
        print ">>>[app onSize] win wasn't open: lets quit!"
        return
      wp=WindowProperties()
      wp.addProperties(wp0)
      if wx.Platform == "__WXMSW__": wp.setOrigin(0,0)
      wp.setSize(*self.win.getP3DSurfaceSize())
      wp.setForeground(True)
      base.win.requestProperties(wp)
      self._kbsync()
    else:
      print ">>>[app onSize] base.win missing!"
    if event != None: event.Skip()
  #------------------------------------------------------------
  #
  def p3dSurfaceFocus(self, surface=1):
    '''lend the focus to the selected p3d window (surface)
    crucial to reroute keyboard events to it
    \note since that onSize() do the job we call it then
    '''
    self.onSize()
  #------------------------------------------------------------
  #
  def autoLoadScript(self):
    '''To autoload a script from commandline using the -a option.
    '''
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option("-a", "--autoload",
      dest="aloadscript", help="Autoload a python script")
    (options,args) = parser.parse_args()
    #
    if options.aloadscript:
      self.appendPea(os.path.abspath(options.aloadscript))
  #------------------------------------------------------------
  #
  def _EVTEST(self, event=None, t="?"):
    print repr(time.time())+" "+t
    if event <> None: event.Skip()
  #------------------------------------------------------------
  #
  def onActivate(self, evt=None):
    '''
    force focus to p3d surface expec. to get keyboard control back
    '''
    if self.vetoActivate:
      evt.Skip()
      return
    if evt.GetActive():
      #print (">>>[onActivate] win ACTIVE")
      self.p3dSurfaceFocus()
      evt.Skip()
    else:
      #print (">>>[onActivate] win IN-ACTIVE")
      evt.StopPropagation()
  #------------------------------------------------------------
  #
  def onClose(self, event):
    while self.wxevt_loop.Pending(): self.wxevt_loop.Dispatch()
    self.win.Destroy()
    try: base.userExit()
    except: sys.exit()
  #------------------------------------------------------------
  #
  def _createMenuBar(self):
    '''Create the main menu bar
    Each main item on the bar gotta related member, which must be used to eventually extend its submenu in the onExtendMenuBar overrided method of the subclass. See sample1 to see how.
    '''
    #--------------------
    # FILE MENUBAR
    self.menuFile = wx.Menu()
    self.win.menuBar.Append(self.menuFile, "&File")
    # open script
    menuItem = self.insertMenuIconItem(
      self.menuFile, 0, self.ID_FILE_OPENSCPT, "&Open pea...", "Open a pea script cleaning up eventually the previous loaded.", wx.ART_FILE_OPEN )
    self.Bind(wx.EVT_MENU, self._mnuOpenScript, menuItem)
    # quit pod
    menuItem = self.menuFile.Append(wx.ID_EXIT, "&Quit")
    self.win.Bind(wx.EVT_MENU, self.onClose, menuItem)
    #--------------------
    # HELP MENUBAR
    self.menuHelp = wx.Menu()
    self.win.menuBar.Append(self.menuHelp, "&Help")
    menuItem = self.menuHelp.Append(wx.ID_ABOUT, "&About...")
    self.win.Bind(wx.EVT_MENU, self._showAbout, menuItem)
    #--------------------
    # CAMERA MENUBAR
    self.menuCamera = wx.Menu()
    self.win.menuBar.Insert(1,self.menuCamera, "&Camera")
    ##
    menuItem = self.menuCamera.Append(self.ID_CAMERA_FPS, "&Toggle FPS", "Toggle fps display")
    self.Bind(wx.EVT_MENU, self.onToggleFPS, menuItem)
    ##
    menuItem = self.menuCamera.Append(self.ID_CAMERA_AA, "&Toggle Antialiasing", "Toggle applying AA to the scene")
    self.Bind(wx.EVT_MENU, self.onToggleAA, menuItem)
    ##
    menuItem = self.menuCamera.Append(
      self.ID_CAMERA_TX, "&Toggle Textures",
      "Toggle textures to all models in the scene"
    )
    self.Bind(wx.EVT_MENU, self.onToggleTextures, menuItem)
    ##
    menuItem = self.menuCamera.Append(
      self.ID_CAMERA_WF, "&Toggle Wireframe",
      "Toggle wireframe visualization to all models in the scene"
    )
    self.Bind(wx.EVT_MENU, self.onToggleWire, menuItem)

    #--------------------
    # custom user extentions of the menu
    self.onExtendMenuBar(self.win.menuBar)
  #------------------------------------------------------------
  #
  def insertMenuIconItem(
      self, menu, pos, itemid, label, hint="", iconid=None
  ):
    menuItem = wx.MenuItem(menu, itemid, label, hint)
    if iconid:
      bmp = wx.ArtProvider.GetBitmap(iconid, wx.ART_MENU, (16,16))
      if not bmp.Ok(): self.clearBmp(wx.EmptyBitmap(16,16))
      menuItem.SetBitmap(bmp)
    self.menuFile.InsertItem(pos, menuItem)
    return menuItem
  #------------------------------------------------------------
  #
  def _showAbout(self, event):
    info = wx.AboutDialogInfo()
    if self.APPINFO['appname']: info.Name = self.APPINFO['appname']
    if self.APPINFO['appversion']:
      info.Version = self.APPINFO['appversion']
    if self.APPINFO['copyright']:
      info.Copyright = self.APPINFO['copyright']
    if self.APPINFO['appsummary']:
      info.Description = self.APPINFO['appsummary']
    if self.APPINFO['website']:
      info.SetWebSite(self.APPINFO['website'])
    if self.APPINFO['devs']: info.Developers = self.APPINFO['devs']
    if self.APPINFO['license']:
      info.License = self.APPINFO['license']
    #
    wx.AboutBox(info)
  #------------------------------------------------------------
  #
  def onToggleFPS(self, event=None):
    base.setFrameRateMeter(not base.frameRateMeter)
  #------------------------------------------------------------
  #
  def onToggleAA(self, event=None):
    if base.render.getAntialias(): base.render.clearAntialias()
    else:
      musa=base.win.getFbProperties().getMultisamples()
      base.render.setAntialias(AntialiasAttrib.MMultisample, musa)
    self.win.SetStatusText(('AA' if base.render.getAntialias() else ''), 1)
    ##print "AA:"+repr(base.render.getAntialias())
    ##print 'AA SAMPLES:',base.win.getFbProperties().getMultisamples()
  #------------------------------------------------------------
  #
  def onToggleWire(self, event=None): base.toggleWireframe()
  #------------------------------------------------------------
  #
  def onToggleTextures(self, event=None): base.toggleTexture()
  #------------------------------------------------------------
  #
  def onScreenshot(self):
    '''override to customize your screenshots'''
    print ">>>screenshot!"
    base.screenshot()
  #------------------------------------------------------------
  #
  def setCamera(self, event=None, action=""):
    if action.lower() == 'reset':
      base.mouseInterfaceNode.setPos(0,0,0)
      base.mouseInterfaceNode.setHpr(0,0,0)
      base.enableMouse()
  #------------------------------------------------------------
  #
  def _mnuOpenScript(self, event=None):
    '''
    '''
    if self.pea_agent[0]:
      if not wxut.questionbox(
        "Will be cleaned up the whole scene:\nAre you sure?",
        parent=self.win,
      ): return
    self.vetoActivate=True
    self.openPea()
    self.vetoActivate=False
  #------------------------------------------------------------
  #
  def openPea(self, filename=""):
    '''
    Open end execute a python script cleaning up first all the previously loaded running scripts.
    If filename is empty a file dialog will appears
    '''
    self.sceneCleanup()
    if (not filename):
      filename=wxut.openFileDialog(parent=self.win,
        message="Choose a Pea Script",
        wildcard="*.py"
      )
    self.appendPea(filename)
  #------------------------------------------------------------
  #
  def sceneCleanup(self):
    # self.pea_agent[0] ! none indicate that we got a loaded pea
    if self.pea_agent[0]:
      if hasattr(self.pea_agent[0], '_final'):
        self.pea_agent[0]._final()
      self.pea_agent[1].restore()
      self.pea_agent[0]=None
      self.setCamera(action='reset')
      return True
    return False
  #------------------------------------------------------------
  #
  def appendPea(self, filename):
    '''Append and execute the p3d python filename, specifiyng also a default prc file structure starting from the script root folder
    '''
    if os.path.isfile(filename):
      #set default .prc common paths for the script
      #NB: font and shaders are searched along duplicates model-path keys
      main_dir=os.path.abspath(os.path.dirname(filename))
      loadPrcFileData("", '''#
model-path $MAIN_DIR/data/models/
model-path $MAIN_DIR/data/fonts/
model-path $MAIN_DIR/data/shaders/
sound-path $MAIN_DIR/data/sounds/
texture-path $MAIN_DIR/data/textures/
particle-path $MAIN_DIR/data/particles/
      '''.replace("$MAIN_DIR", main_dir))
      sys.path.append(main_dir)
      splash=splashCard('loading.png')
      #open and execute a pea
      try:
        #
        sg=globals()
        #to be used by the pea just in case there is need to load custom resources with python I/O functions
        sg['_PEA_ROOT']=main_dir
        execfile(filename, sg)
        #\note pea global must exists in the loaded script as the main class of the app
        if 'pea' in globals():
          self.pea_agent[0]=pea
        else:
          self.pea_agent[0]=None
          print "[appendPea] No pea found!"
        #
        self.win.SetStatusText(
          "Pea '%s' loaded" % os.path.basename(filename)
        )
      except:
        print "[ERROR:appendpea]file:%s"%filename
        import traceback
        print traceback.format_exc(1)
      splash.destroy()
    else:
      self.win.SetStatusText("Script loading canceled")
  #------------------------------------------------------------
  # LITL HELPERS
  #------------------------------------------------------------
  #
  def getconfstring(self, cvar, default=''):
    '''
    read from the p3d configuration .prc files
    returns the string representation of the specified config var
    '''
    v=ConfigVariableString(cvar, default)
    return v.getValue()
  #------------------------------------------------------------
  #
  def getconfpath(self, cvar, default=''):
    '''
    read from the p3d configuration .prc files
    returns a list representing the specified config var paths
    \note the [0] index is the last added, possibly the one specified in the custom .prc for p3dwxpod but maybe another added by you.
    '''
    v=ConfigVariableSearchPath(cvar, default)
    return v.__str__().split('\n')
  #------------------------------------------------------------
  # USER METHODS TO OVERRIDE
  #------------------------------------------------------------
  #
  def onStartP3D(self):
    '''override in a child class to put initialization stuff
    '''
    pass
  #------------------------------------------------------------
  #
  def onExtendMenuBar(self, mainmenu):
    '''override in a child class to extend the menubar
    '''
    pass
#========================================================================
#
from pandac.PandaModules import TransparencyAttrib
from direct.gui.OnscreenImage import OnscreenImage
class splashCard(object):
  #------------------------------------------------------------
  #
  def __init__(self, image):
    self.myImage=OnscreenImage(image=image, pos = (0, 0, 0),
      parent=base.render2d, scale=(.3,1.,.3)
    )
    self.myImage.setTransparency(TransparencyAttrib.MAlpha)
    #let it shows
    for i in range(4):
      base.graphicsEngine.renderFrame()
      time.sleep(0.3)
  #------------------------------------------------------------
  #
  def destroy(self): self.myImage.destroy()
#========================================================================
#
if __name__ == "__main__":
  import wxutils as wxut
  from direct.actor import Actor
  from pandac.PandaModules import DirectionalLight, VBase4
  class MyApp(P3dWxApp):
    WIDTH=540
    HEIGHT=420
    ID_MODEL_LOADM=1010
    ID_MODEL_LOADA=1020
    #------------------------------------------------------------------
    #
    def onStartP3D(self):
      self.APPINFO['appname']='A basic sample app'
      self.APPINFO['appversion']="wip.2009-05"
      self.APPINFO['appsummary']="Minimal setup to start with p3dwxpod"
      self.APPINFO['copyright']="(c) 2009 Fabius Astelix"
      self.APPINFO['license']="Free for everybody"

      dlight = DirectionalLight('dlight')
      dlight.setColor(VBase4(0.9, 0.9, 0.9, 1))
      dlnp = render.attachNewNode(dlight)
      dlnp.setHpr(0, -60, 0)
      base.render.setLight(dlnp)

      base.mouseInterfaceNode.setPos(0, 20, -4.5)
    #------------------------------------------------------------------
    #
    def onExtendMenuBar(self, mainmenu):
      #--------------------
      # extend the mainmenu here
      #--------------------
      # Add a model menubar
      self.menuModel = wx.Menu()
      mainmenu.Insert(1, self.menuModel, "&Model")
      ##
      menuItem = self.menuModel.Append(self.ID_MODEL_LOADM, "&Load Model...", "Open the file dialog to load a model.")
      self.Bind(wx.EVT_MENU, self.mnuLoadModel, menuItem)
    #------------------------------------------------------------------
    #
    def mnuLoadModel(self, event=None):
      egg=self.loadEgg(start=self.getconfpath('model-path')[0])
      if egg:
        splash=splashCard('loading.png')
        ob=loader.loadModel(egg)
        self.win.SetStatusText("%s loaded." % os.path.basename(egg))
        ob.reparentTo(base.render)
        #ob.setScale(0.25,0.25,0.25)
        ob.setPos(0, 12, 0)
        splash.destroy()
    #------------------------------------------------------------------
    #
    def loadEgg(self, kind="Model", start=""):
      eggfile=wxut.openFileDialog(parent=self.win,
        message="Choose "+kind,
        start=start,
        wildcard="*.egg;*.egg.pz"
      )
      if os.path.isfile(eggfile):
        self.win.SetStatusText("Loading '%s'..." % os.path.basename(eggfile))
        return eggfile
      else:
        self.win.SetStatusText(kind+" Load Canceled")
      return None

  _app=MyApp(0)
  run()
