'''
by: Fabius Astelix
license: Free for everybody
summary: useful stuff to use in P3D scripts

\fixme
-le light non vengono pulite bene
'''
from pandac.PandaModules import WindowProperties, loadPrcFile, ConfigVariableSearchPath, ConfigVariableString, loadPrcFileData
loadPrcFileData("", 'want-dev #t')
from pandac.PandaModules import MemoryUsage
from direct.gui.OnscreenText import OnscreenText
from direct.task import Task
# ivalMgr IntervalManager is here
from direct.interval.IntervalGlobal import *
import copy, sys
#===================================================================
#
class PwxResmanager(object):
  #---------------------------------------------------------
  def __init__(self, globs, setwindow=None, debug=False):
    self.reopenwindow=setwindow
    self._client_globals=globs
    self.debug=debug
    self.preserve()
  #---------------------------------------------------------
  def scvsize(self, sz):
    '''convert a size to the byte-size weight multiple close to 0'''
    scale = [('GB', 1024.*1024.*1024.), ('MB', 1024.*1024.), ('KB', 1024.),]
    for x in scale:
      dx=(sz/x[1])
      if dx > 1.: break
    return "%0.2f%s"%(dx, x[0])
  #-----------------------------------------
  def get_memory(self):
    return self.scvsize(MemoryUsage.getTotalSize())
  #---------------------------------------------------------
  def restore(self):
    '''restoring to render default situation'''
    self.printdbg(">>>>Scene Restoration")
    self.restore_sounds()
    self.printdbg(">>sounds restored")
    self.restore_tasks()
    self.printdbg(">>tsk restored")
    self.restore_lights()
    self.printdbg(">>lights restored")
    self.restore_dgui()
    self.printdbg(">>dgui restored")
    #restore render2d tree
    self.restore_r2d()
    #restore render tree and his cam
    base.render.node().removeAllChildren()
    base.render.node().stealChildren(self._default_scene)
    base.camera=base.render.find("**/camera")
    base.render.find("**/camera/cam").removeNode()
    base.cam=base.camNode=base.camLens=None
    base.dataUnused.node().removeAllChildren()
    #call window restart handler
    if self.reopenwindow: self.reopenwindow()
    #reset trackball mode
    base.setMouseOnNode(base.camera.node())
    base.useTrackball()
    #
    self._default_scene=base.render.node().copySubgraph()
    base.setBackgroundColor(self._default_bgcolor)
    if self._default_fog: base.render.setFog(self._default_fog)
    else: base.render.clearFog()
    self.printdbg(">>render restored")
    self.restore_globals()
    self.printdbg(">>globals restored")
    self.printdbg(">>>>end scene restore")
  #---------------------------------------------------------
  def restore_r2d(self):
    all=base.render2d.findAllMatches("**/*")
    for i in range(all.getNumPaths()):
      if not self._default_r2d.hasPath(all[i]):
        all[i].removeNode()
  #---------------------------------------------------------
  def restore_globals(self):
    glo=copy.copy(self._client_globals)
    for g in [
      gi for gi in glo.items() if gi[0] not in self._default_globals
    ]:
      if not g[0].startswith('_'):
        del self._client_globals[g[0]]
        self.printdbg("%r [%s] (deglobized)" % (g[0], type(g[1])))
  #---------------------------------------------------------
  def restore_tasks(self):
    for t in taskMgr.getTasks():
      if t.id not in self._default_tasks: t.remove()

    for i in ivalMgr.getIntervalsMatching('*'):
      if not i in self._default_intvl:
        ivalMgr.removeInterval(i)

    ###for e in base.messenger.getEvents():
      ###obs=base.messenger.whoAccepts(e).keys()
      ###for o in obs:
        ###if  (e not in self._default_events) or \
            ###(o not in self._default_events[e]):
          ###base.messenger.ignore(e, o)
    base.messenger.clear()
    for e,m in self._default_events.iteritems():
      base.accept(e, m)
  #---------------------------------------------------------
  def restore_lights(self):
    all=base.render.findAllMatches("**/+Light")
    for i in range(all.getNumPaths()):
      print ">>>l:"+all[i].getName()
      if not all[i].getName().startswith("_"):
        base.render.clearLight(all[i])
  #---------------------------------------------------------
  def restore_dgui(self):
    dgoobjs=[
      item for item in base.guiItems.values()
        if not item in self._default_dgui
      ]
    dgopaths=[repr(dgo) for dgo in dgoobjs]
    dguistuff=base.render2d.findAllMatches("**/+PGItem")
    for i in range(dguistuff.getNumPaths()):
      try:
        idx=dgoobjs.index(dguistuff[i])
        dgoobjs[idx].destroy()
        dguistuff[i].removeNode()
      except:
        import traceback
        self.printdbg("[ERROR]%s"%repr(dguistuff[i]))
        self.printdbg(traceback.format_exc(1))
  #---------------------------------------------------------
  def restore_sounds(self):
    for am in base.sfxManagerList:
      if self.debug: print "SND>>"+repr(am)+" stopped"
      am.stopAllSounds()
    base.musicManager.stopAllSounds()
  #---------------------------------------------------------
  def info(self):
    if self.debug:
      print "\n"+"#"*60
      print "\n>>render"
      base.render.ls()
      print "\n>>render2d"
      base.render2d.ls()
      print "\n>>dataroot"
      base.dataRoot.ls(ostream,2)
      print "\n>>dgui"
      dguistuff=base.render2d.findAllMatches("**/+PGItem")
      for i in range(dguistuff.getNumPaths()):
        if not dguistuff[i] in self._default_dgui:
          self.printdbg(dguistuff[i].getName()+" (sweepable)")
        else:
          self.printdbg(dguistuff[i].getName()+" (untouchable)")
      print "\n>>actor"
      act=base.render.findAllMatches("**/+Character")
      for i in range(act.getNumPaths()):
        print act[i].getName()+""
      print "\n>>tasks"
      for t in base.taskMgr.getTasks():
        if t.id not in self._default_tasks: print t
      print ivalMgr.getIntervalsMatching('*')
      print "\n>>messenger"
      ###for e in base.messenger.getEvents():
        ###print "%r::%r"%(e, base.messenger.whoAccepts(e).keys())
      for e in base.messenger.getEvents():
        print "~"*5
        print "...who(%r)?%r"%(
          e,
          base.messenger.whoAccepts(e).values()[0][0]
        )
      ###print "\n>>drs"
      ###for ndr in range(base.win.getNumDisplayRegions()):
        ###print "%2d) %r"%(ndr, base.win.getDisplayRegion(ndr))
      ###print "\n>>geng"
      ###for i in range(base.graphicsEngine.getNumWindows()):
        ###w=base.graphicsEngine.getWindow(i)
        ###if isinstance(w, ParasiteBuffer):
          ###print repr(w)+"::"+" you shall die!!!"
        ###else:
          ###print repr(w)+"::"+" you stay"
  #---------------------------------------------------------
  def printdbg(self, s):
    if self.debug: print s
  #---------------------------------------------------------
  def preserve(self):
    #render subtree preservation
    self._default_scene=base.render.node().copySubgraph()
    self._default_r2d=base.render2d.findAllMatches("**/*")
    self._default_tasks=[t.id for t in taskMgr.getTasks()]
    self._default_dgui=base.guiItems.values()
    self._default_intvl=ivalMgr.getIntervalsMatching('*')
    #lights with names like '_name' are automatically preserved
    self._default_globals=copy.copy(self._client_globals.keys())
    self._default_bgcolor=base.getBackgroundColor()
    self._default_fog=base.render.getFog()
    self._default_events={}
    for e in base.messenger.getEvents():
      self._default_events[e]=base.messenger.whoAccepts(e).values()[0][0]
#===================================================================
#
class track_pos(object):
  '''
  helper object to show the position of a node on the screen
  \todo dare nomi univoci agli oggetti-testo e autoposizionarli cercando nel nodegraph col match sui nomi
  '''
  def __init__(self, subject=None, textpos=(-1, -.97), fgcolor=(1,1,1,1)):
    self._subject=subject or base.camera
    self._display=OnscreenText(
      text='',
      pos = (textpos[0], textpos[1]),
      fg=fgcolor,
      mayChange = True,
      scale=0.05,
    )
    taskMgr.add(self.tsk_track,'track_pos')

  def tsk_track(self, task):
    pos=self._subject.getPos()
    self._display.setText(
      "%s pos : %3.3f %3.3f %3.3f" % (
        self._subject.getName().title(), pos[0], pos[1], pos[2]
      )
    )

    return Task.cont
#========================================================================
# FUNCTIONS
#========================================================================
def prcpath(cvar, default=''):
  '''pick a path value from the p3d configuration pool'''
  v=ConfigVariableSearchPath(cvar, default)
  return v.__str__().split('\n')[0]
